@@ -44,7 +44,6 @@
   /* `stdin' file descriptor is not in line-by-line input mode.       */
 #ifdef _WIN32
 #include <conio.h>
-#define snprintf  _snprintf
 #define getch     _getch
 #else
 #define getch     getchar
@@ -53,7 +52,6 @@
 
 #include <ft2build.h>
 #include <freetype/freetype.h>
-
 #include <freetype/ftdriver.h>
 #include <freetype/ftmm.h>
 
@@ -61,33 +59,26 @@
 #include "strbuf.h"
 #include "mlgetopt.h"
 
-
-  /* The following header shouldn't be used in normal programs.    */
+  /* The following headers shouldn't be used in normal programs.  */
   /* `freetype/src/truetype' must be in the current include path. */
-#include "ttobjs.h"
-#include "ttdriver.h"
 #include "ttinterp.h"
 #include "tterrors.h"
 
 
 #define Quit     -1
 #define Restart  -2
+#define Reload   -3
 
 
   static FT_Library    library;    /* root library object */
-  static FT_Memory     memory;     /* system object       */
-  static FT_Driver     driver;     /* truetype driver     */
-  static TT_Face       face;       /* truetype face       */
-  static TT_Size       size;       /* truetype size       */
-  static TT_GlyphSlot  glyph;      /* truetype glyph slot */
+  static FT_Face       face;       /* truetype face       */
 
-  static FT_MM_Var    *multimaster;
   static FT_Fixed*     requested_pos;
   static unsigned int  requested_cnt;
 
-  static unsigned int  tt_interpreter_versions[2];
-  static int           num_tt_interpreter_versions;
-  static unsigned int  dflt_tt_interpreter_version;
+  static unsigned int  versions[] = { 0,  /* reserved for active */
+                                      TT_INTERPRETER_VERSION_35,
+                                      TT_INTERPRETER_VERSION_40 };
 
   /* number formats */
   enum {
@@ -1545,16 +1536,22 @@
   static void
   Abort( const char*  message )
   {
-    const FT_String  *str;
+    if ( error != Quit )
+    {
+      const FT_String  *str;
 
 
-    switch( error )
-    #include <freetype/fterrors.h>
+      switch( error )
+      #include <freetype/fterrors.h>
 
-    fprintf( stderr, "%s\n  error = 0x%04x, %s\n", message, error, str );
+      fprintf( stderr, "%s\n  error = 0x%04x, %s\n", message, error, str );
+    }
 
     Reset_Keyboard();
-    exit( 1 );
+    free( requested_pos );
+    FT_Done_FreeType( library );
+
+    exit( error != Quit );
   }
 
 
@@ -1563,32 +1560,34 @@
   {
     unsigned int  i;
     char*         s;
+    char*         t;
 
 
-    /* get number of coordinates;                                      */
-    /* a group of non-whitespace characters is handled as one argument */
+    /* get number of coordinates */
     s = arg;
-    for ( requested_cnt = 0; *s; requested_cnt++ )
+    for ( i = 0; *s; i++, s = t )
     {
-      while ( isspace( *s ) )
-        s++;
+      strtod( s, &t );  /* idle parsing */
 
-      while ( *s && !isspace( *s ) )
+      if ( t == s )
+        break;
+      s = t;
+
+      if ( *s == ',' )
         s++;
     }
 
-    requested_pos = (FT_Fixed*)malloc( sizeof ( FT_Fixed ) * requested_cnt );
+    requested_pos = (FT_Fixed*)malloc( sizeof ( FT_Fixed ) * i );
+    if ( !requested_pos )
+      return;
+    requested_cnt = i;
 
     s = arg;
     for ( i = 0; i < requested_cnt; i++ )
     {
       requested_pos[i] = (FT_Fixed)( strtod( s, &s ) * 65536.0 );
-      /* skip until next whitespace in case of junk */
-      /* that `strtod' doesn't handle               */
-      while ( *s && !isspace( *s ) )
-        s++;
 
-      while ( isspace( *s ) )
+      if ( *s == ',' )
         s++;
     }
   }
@@ -2061,8 +2060,11 @@
       code_range = "fpgm";
     }
 
-    printf( "Entering `%s' table.\n"
-            "\n", code_range );
+    printf( "\n"
+            "Entering `%s' table.\n"
+            "Twilight points: %hu. Glyph points: %hu%s.\n"
+            "\n", code_range, twilight.n_points, pts.n_points,
+                  exc->is_composite ? ", composite" : "" );
 
     really_leave = 0;
 
@@ -2205,10 +2207,7 @@
           }
         }
         else
-        {
-          printf( "\n" );
           goto LErrorLabel_;
-        }
       }
 
       if ( breakpoint.IP == CUR.IP          &&
@@ -2229,7 +2228,7 @@
           printf(
             "ttdebug Help\n"
             "\n"
-            "Q   quit debugger                         V   show vector info\n"
+            "L   reload glyph                          V   show vector info\n"
             "R   restart debugger                      G   show graphics state\n"
             "c   continue to next code range           P   show points zone\n"
             "n   skip to next instruction              T   show twilight zone\n"
@@ -2240,7 +2239,8 @@
             "p   toggle breakpoint at prev. position   O   show opcode docstring\n"
             "F   cycle value format (int, float, 64th)\n"
             "I   toggle hex/decimal integer format     H   show format help\n"
-            "\n" );
+            "Q   quit debugger\n"
+            "\n");
           break;
 
         case 'H':
@@ -2606,6 +2606,14 @@
         error = Restart;
         goto LErrorLabel_;
 
+      /* reload glyph */
+      case 'L':
+        /* without the pedantic hinting flag,                   */
+        /* FreeType ignores bytecode errors in `glyf' programs  */
+        CUR.pedantic_hinting = 1;
+        error = Reload;
+        goto LErrorLabel_;
+
       /* continue */
       case 'c':
         if ( CUR.IP < CUR.codeSize )
@@ -2848,7 +2856,7 @@
     free( storage );
     free( save_storage );
 
-    if ( error && error != Quit && error != Restart )
+    if ( error > 0 )
       Abort( "error during execution" );
 
     return error;
@@ -2858,19 +2866,30 @@
   static void
   Usage( const char*  execname )
   {
-    char  versions[32];
+    int     last = sizeof ( versions ) / sizeof ( versions[0] );
+    int     i;
+    char    *dlm = "";
+    char    buf[32];
+    StrBuf  sb[1];
 
 
-    /* we expect that at least one interpreter version is available */
-    if ( num_tt_interpreter_versions == 1 )
-      snprintf( versions, sizeof ( versions ),
-                "%d",
-                tt_interpreter_versions[0] );
-    else
-      snprintf( versions, sizeof ( versions ),
-                "%d and %d",
-                tt_interpreter_versions[0],
-                tt_interpreter_versions[1] );
+    strbuf_init( sb, buf, sizeof ( buf ) );
+    strbuf_reset( sb );
+
+    /* check available versions */
+    for ( i = 1; i < last; i++ )
+    {
+      error = FT_Property_Set( library,
+                               "truetype",
+                               "interpreter-version", versions + i );
+      if ( !error )
+      {
+        strbuf_format( sb, "%s%d%s",
+                       dlm, versions[i],
+                       versions[i] == versions[0] ? " (active)" : "");
+        dlm = ", ";
+      }
+    }
 
     fprintf( stderr,
       "\n"
@@ -2886,27 +2905,21 @@
       "  font      The TrueType font file to debug.\n"
       "\n"
       "  -I ver    Use TrueType interpreter version VER.\n"
-      "            Available versions are %s; default is version %d.\n"
+      "            Available versions: %s.\n"
       "  -f idx    Access font IDX if input file is a TTC (default: 0).\n"
-      "  -d \"axis1 axis2 ...\"\n"
+      "  -a axis1,axis2,...\n"
       "            Specify the design coordinates for each variation axis\n"
       "            at start-up (ignored if not a variation font).\n"
       "  -v        Show version.\n"
       "\n"
       "While running, press the `?' key for help.\n"
       "\n",
-      versions,
-      dflt_tt_interpreter_version );
+      buf );
 
     exit( 1 );
   }
 
 
-  static char*         file_name;
-  static unsigned int  glyph_index;
-  static int           glyph_size;
-
-
   int
   main( int     argc,
         char**  argv )
@@ -2914,63 +2927,50 @@
     int    option;
     char   version_string[64];
 
-    int           i;
     const char*   execname;
-    unsigned int  versions[2] = { TT_INTERPRETER_VERSION_35,
-                                  TT_INTERPRETER_VERSION_40 };
     int           version;
     int           face_index = 0;
+    FT_UInt       glyph_index;
+    FT_UInt       glyph_size;
+    FT_MM_Var    *multimaster;
 
-    int  tmp;
+    FT_Open_Args  args = { FT_OPEN_PATHNAME | FT_OPEN_DRIVER };
 
 
-    /* init library, read face object, get driver, create size */
+    /* init library */
     error = FT_Init_FreeType( &library );
     if ( error )
       Abort( "could not initialize FreeType library" );
 
-    memory = library->memory;
-    driver = (FT_Driver)FT_Get_Module( library, "truetype" );
-    if ( !driver )
-      Abort( "could not find the TrueType driver in FreeType 2\n" );
+    args.driver = FT_Get_Module( library, "truetype" );
+    if ( !args.driver )
+      Abort( "could not find the TrueType driver in FreeType\n" );
+
+    /* get the default interpreter version */
+    FT_Property_Get( library,
+                     "truetype",
+                     "interpreter-version", versions );
 
     {
       FT_Int  major, minor, patch;
-      int     offset;
+      StrBuf  sb[1];
 
 
+      strbuf_init( sb, version_string, sizeof ( version_string ) );
+      strbuf_reset( sb );
+
       FT_Library_Version( library, &major, &minor, &patch );
 
-      offset = snprintf( version_string, 64,
-                         "ttdebug (FreeType) %d.%d",
-                         major, minor );
+      strbuf_format( sb, "ttdebug (FreeType) %d.%d", major, minor );
       if ( patch )
-        offset = snprintf( version_string + offset, (size_t)( 64 - offset ),
-                           ".%d",
-                           patch );
+        strbuf_format( sb, ".%d", patch );
     }
 
-    /* collect all available versions, then set again the default */
-    FT_Property_Get( library,
-                     "truetype",
-                     "interpreter-version", &dflt_tt_interpreter_version );
-    for ( i = 0; i < 2; i++ )
-    {
-      error = FT_Property_Set( library,
-                               "truetype",
-                               "interpreter-version", &versions[i] );
-      if ( !error )
-        tt_interpreter_versions[num_tt_interpreter_versions++] = versions[i];
-    }
-    FT_Property_Set( library,
-                     "truetype",
-                     "interpreter-version", &dflt_tt_interpreter_version );
-
     execname = ft_basename( argv[0] );
 
     while ( 1 )
     {
-      option = getopt( argc, argv, "I:d:f:v" );
+      option = getopt( argc, argv, "I:a:f:v" );
 
       if ( option == -1 )
         break;
@@ -2980,31 +2980,19 @@
       case 'I':
         version = atoi( optarg );
 
-        if ( version < 0 )
-        {
-          printf( "invalid TrueType interpreter version = %d\n", version );
-          Usage( execname );
-        }
+        error = FT_Property_Set( library,
+                                 "truetype",
+                                 "interpreter-version", &version );
 
-        for ( i = 0; i < num_tt_interpreter_versions; i++ )
-        {
-          if ( (unsigned int)version == tt_interpreter_versions[i] )
-          {
-            FT_Property_Set( library,
-                             "truetype",
-                             "interpreter-version", &version );
-            break;
-          }
-        }
-
-        if ( i == num_tt_interpreter_versions )
+        if ( error )
         {
           printf( "invalid TrueType interpreter version = %d\n", version );
           Usage( execname );
         }
+        versions[0] = version;
         break;
 
-      case 'd':
+      case 'a':
         parse_design_coords( optarg );
         break;
 
@@ -3030,22 +3018,50 @@
       Usage( execname );
 
     /* get glyph index */
-    if ( sscanf( argv[0], "%d", &tmp ) != 1 || tmp < 0 )
+    if ( sscanf( argv[0], "%u", &glyph_index ) != 1 )
     {
-      printf( "invalid glyph index = %s\n", argv[1] );
+      printf( "invalid glyph index = %s\n", argv[0] );
       Usage( execname );
     }
-    glyph_index = (unsigned int)tmp;
 
     /* get glyph size */
-    if ( sscanf( argv[1], "%d", &glyph_size ) != 1 || glyph_size < 0 )
+    if ( sscanf( argv[1], "%u", &glyph_size ) != 1 )
     {
       printf( "invalid glyph size = %s\n", argv[1] );
       Usage( execname );
     }
 
     /* get file name */
-    file_name = argv[2];
+    args.pathname = argv[2];
+
+    error = FT_Open_Face( library, &args, face_index, &face );
+    if ( error || !face->num_glyphs )
+      Abort( "could not open input font file" );
+
+    glyph_index %= face->num_glyphs;
+
+    if ( requested_cnt && !FT_Get_MM_Var( face, &multimaster ) )
+    {
+      unsigned int  n;
+
+
+      if ( requested_cnt > multimaster->num_axis )
+        requested_cnt = multimaster->num_axis;
+
+      for ( n = 0; n < requested_cnt; n++ )
+      {
+        if ( requested_pos[n] < multimaster->axis[n].minimum )
+          requested_pos[n] = multimaster->axis[n].minimum;
+        else if ( requested_pos[n] > multimaster->axis[n].maximum )
+          requested_pos[n] = multimaster->axis[n].maximum;
+      }
+
+      FT_Done_MM_Var( library, multimaster );
+
+      FT_Set_Var_Design_Coordinates( face,
+                                     requested_cnt,
+                                     requested_pos );
+    }
 
     Init_Keyboard();
 
@@ -3053,79 +3069,61 @@
                        FT_DEBUG_HOOK_TRUETYPE,
                        RunIns );
 
-    printf( "%s\n"
-            "press key `h' or `?' for help\n"
-            "\n", version_string );
+    printf( "%s, TrueType v%d\n"
+            "%s %s, %u ppem, glyph %u\n"
+            "press key `h' or `?' for help\n",
+            version_string, versions[0],
+            face->family_name, face->style_name,
+            glyph_size, glyph_index );
 
-    while ( !error )
+    do
     {
-      error = FT_New_Face( library, file_name, face_index, (FT_Face*)&face );
-      if ( error )
-        Abort( "could not open input font file" );
-
-      /* find driver and check format */
-      if ( face->root.driver != driver )
-      {
-        error = FT_Err_Invalid_File_Format;
-        Abort( "this is not a TrueType font" );
-      }
-
-      FT_Done_MM_Var( library, multimaster );
-      error = FT_Get_MM_Var( (FT_Face)face, &multimaster );
-      if ( error )
-        multimaster = NULL;
-      else
-      {
-        unsigned int  n;
-
-
-        if ( requested_cnt > multimaster->num_axis )
-          requested_cnt = multimaster->num_axis;
-
-        for ( n = 0; n < requested_cnt; n++ )
-        {
-          if ( requested_pos[n] < multimaster->axis[n].minimum )
-            requested_pos[n] = multimaster->axis[n].minimum;
-          else if ( requested_pos[n] > multimaster->axis[n].maximum )
-            requested_pos[n] = multimaster->axis[n].maximum;
-        }
-
-        FT_Set_Var_Design_Coordinates( (FT_Face)face,
-                                       requested_cnt,
-                                       requested_pos );
-      }
-
-      size = (TT_Size)face->root.size;
-
-      error = FT_Set_Char_Size( (FT_Face)face,
-                                glyph_size << 6,
-                                glyph_size << 6,
-                                72,
-                                72 );
+      error = FT_Set_Pixel_Sizes( face,
+                                  glyph_size,
+                                  glyph_size );
       if ( error )
         Abort( "could not set character size" );
 
-      glyph = (TT_GlyphSlot)face->root.glyph;
-
       /* now load glyph */
-      error = FT_Load_Glyph( (FT_Face)face,
-                             (FT_UInt)glyph_index,
+      error = FT_Load_Glyph( face,
+                             glyph_index,
                              FT_LOAD_NO_BITMAP );
-      if ( error && error != Quit && error != Restart )
-        Abort( "could not load glyph" );
-      if ( error == Restart )
-        error = FT_Err_Ok;
 
-      FT_Done_Face( (FT_Face)face );
-    }
+      while ( !error )
+      {
+        printf( "\n"
+                "Glyph loaded.\n"
+                "Q to quit, R to restart, L to reload.\n" );
 
-    Reset_Keyboard();
+        switch( getch() )
+        {
+        case 'Q': error = Quit;    break;
+        case 'R': error = Restart; break;
+        case 'L': error = Reload;  break;
+        }
+      }
 
-    FT_Done_FreeType( library );
+      switch ( error )
+      {
+        case Restart:
+          /* rebuild size to destroy bytecode data */
+          FT_Done_Size( face->size );
+          FT_New_Size( face, &face->size );
+          break;
 
-    free( requested_pos );
+        case Reload:
+          break;
 
-    return 0;
+        case Quit:
+          Abort( NULL );
+
+        default:
+          Abort( "could not load glyph" );
+      }
+
+    } while ( 1 );
+
+    return 0;  /* unreachable */
   }
 
 
